/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-client/src/org/opendj/client/ChannelUpdater.java,v 1.1.1.1 2001/09/29 20:22:00 jonathan Exp $
 * $Log: ChannelUpdater.java,v $
 * Revision 1.1.1.1  2001/09/29 20:22:00  jonathan
 * Imported opendj-client
 *
 */
package org.opendj.client;

import javax.swing.table.DefaultTableModel;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import org.opendj.share.Util;
import org.opendj.share.ChannelTime;
import org.opendj.share.ChannelTimeComparator;

public class ChannelUpdater implements Runnable {
    
    // One fifth of a second
    public static final long SLEEP_INTERVAL = 200;
    public static final long REPOPULATE_INTERVAL = 5 * 60; // every minute
    public static final long BROADCAST_REQUEST_MILLIS = (30*1000);
    public static final long BROADCAST_END_MILLIS     = (30*1000);

    private Main itsApp = null;
    private Server itsServer = null;
    private List itsChannelList = null;
    private volatile boolean isDone = false;
    private DefaultTableModel itsModel = null;
    private volatile boolean isRepopulateSuspended = false;
    private volatile boolean itsExitMonitorDone = false;
    private ChannelTime itsActiveChannel = null;
    private Encoder itsEncoder = null;
    private ExitMonitor itsExitMonitor = null;
    private long itsCounter = 0;
    private long itsRealTimeLeft = 0;

    public ChannelUpdater () {}

    public void init ( Main theApp, DefaultTableModel theModel ) {
        itsApp = theApp;
        itsServer = itsApp.getServer();
        itsModel = theModel;
        isDone = false;
        isRepopulateSuspended = false;
        itsExitMonitorDone = false;
        itsCounter = 0;
        itsEncoder = null;
        itsExitMonitor = null;
        itsChannelList = null;
    }

    public void finish() {
        isDone = true;
        isRepopulateSuspended = true;
        if ( itsEncoder != null ) {
            itsActiveChannel = null;
            itsEncoder.finish();
        }
    }

    public void run () {

        populateChannelTable();
        while ( !isDone ) {

            // Sleep a bit
            Util.sleep(SLEEP_INTERVAL);

            // Every once in a while, get fresh data from the server
            if ( ((++itsCounter) % REPOPULATE_INTERVAL) == 0 ) {
                if ( !isRepopulateSuspended ) {
                    populateChannelTable();
                }
                itsApp.refreshShowStats();                
                continue;
            }

            recalculateTimes();
        }
        
        clearTable();
    }

    public void exitMonitorDone() { 
        forceRepopulate();
        itsExitMonitorDone = true;
    }

    private void recalculateTimes () {
        if ( itsChannelList == null ) return;
        Iterator i = itsChannelList.iterator();
        ChannelTime aChannelTime = null;
        int aRowNum = 0;
        while ( i.hasNext() ) {
            aChannelTime = (ChannelTime) i.next();
            itsModel.setValueAt(formatChannelTime(aChannelTime), aRowNum++, 1);
        }
    }

    private void clearTable () {
        // Erase all rows
        itsModel.setNumRows(0);
    }

    private void forceRepopulate () {
        isRepopulateSuspended = false;
        itsCounter = -1;
    }

    private void setRepopulateInMillis ( long theMillis ) {
        isRepopulateSuspended = false;
        itsCounter = (long) (-1.0 * (((double) theMillis)/((double) SLEEP_INTERVAL)));
    }

    private void populateChannelTable() {

        clearTable();

        // If the ExitMonitor finished, clear the ExitMonitor
        // and Encoder fields, and switch the flag back
        if ( itsExitMonitorDone ) {
            itsExitMonitor = null;
            itsEncoder = null;
            itsExitMonitorDone = false;
        }

        // Get new rows from server
        List someTimes = null;
        try {
            someTimes = itsServer.getChannelTimes();
        } catch ( ServerException se ) {
            AlertDialog a = new AlertDialog(itsApp,
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ERROR_GETCHANNELTIMES_FAILED,
                                            se.getMessage());
            return;
            
        } catch ( Exception ex ) {
            AlertDialog a = new AlertDialog(itsApp,
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ERROR_GETCHANNELTIMES_FAILED,
                                            ex);
            return;
        }

        // They have no times
        if ( someTimes == null ) return;
        
        // Iterate over the rows, and build a map that maps 
        // channel names to the soonest (smallest delta) 
        // broadcast time
        Map aChannelMap = new HashMap();
        Iterator i = someTimes.iterator();
        ChannelTime aChannelTime = null;
        ChannelTime aMappedTime = null;
        while ( i.hasNext() ) {
            aChannelTime = (ChannelTime) i.next();
            log("Found ChannelTime: " + aChannelTime);
            aMappedTime = (ChannelTime) aChannelMap.get(aChannelTime.name);
            if ( aMappedTime == null ) {
                aChannelMap.put(aChannelTime.name, aChannelTime);
            } else {
                // Special values of ONAIR supercede all others
                if ( aMappedTime.special == ChannelTime.ONAIR ) {
                    continue;
                }
                if ( aChannelTime.special == ChannelTime.ONAIR ) {
                    aChannelMap.put(aChannelTime.name, aChannelTime);
                    continue;
                }

                // Finally, if this is actually earlier than the mapped,
                // update the map
                if ( aChannelTime.delta < aMappedTime.delta ) {
                    aChannelMap.put(aChannelTime.name, aChannelTime);                    
                }
            }
        }

        // Now that we have a Map, let's iterate over all the keys 
        // and stick them into a vector
        itsChannelList = new Vector();
        i = aChannelMap.keySet().iterator();
        Object key = null;
        while ( i.hasNext() ) {
            key = i.next();
            itsChannelList.add(aChannelMap.get(key));
        }
        
        // Sort the list
        Collections.sort(itsChannelList, new ChannelTimeComparator());

        // NOW, we can add each item in the list to our table
        i = itsChannelList.iterator();
        String[] aRowData = { Util.EMPTY_STRING, Util.EMPTY_STRING };
        while ( i.hasNext() ) {
            aChannelTime = (ChannelTime) i.next();
            aRowData[0] = aChannelTime.name;
            aRowData[1] = formatChannelTime(aChannelTime);
            itsModel.addRow(aRowData);
        }
    }

    private String formatChannelTime( ChannelTime theChannelTime ) {

        String rstr = "";

        long now = System.currentTimeMillis();
        long aRealDelta = theChannelTime.delta - (now - theChannelTime.recvTime);

        // Take care of the urgent cases
        if ( theChannelTime.special == ChannelTime.ONAIR ) {
            // For ONAIR, display the length of time remaining in the broadcast

            // In ONAIR cases, the delta actually indicates how long ago (in a negative number)
            // they should have STARTED broadcasting.  In these cases, we want to display
            // how much longer they have.  Plus, we have to adjust the RealDelta to make it
            // more negative by whatever difference exists between now and the receive time.
            long aTimeLeft = aRealDelta + theChannelTime.length;

            // If we are getting close to the end time, tell the encoder to wrap up
            // log("Checking aTimeLeft=" + aTimeLeft + ", BROADCAST_END_MILLIS=" + BROADCAST_END_MILLIS);
            if ( aTimeLeft > 0 && aTimeLeft < BROADCAST_END_MILLIS ) {
                if ( itsEncoder != null && itsExitMonitor == null ) {
                    // Create an exit monitor to handle 
                    // the termination, and let us know when things have
                    // ended
                    log("Starting ExitMonitor...");
                    itsExitMonitor = new ExitMonitor(itsApp, this, itsEncoder);
                    Thread anExitThread = new Thread(itsExitMonitor);
                    anExitThread.start();
                }
            }

            // If the timeleft is negative, change the message
            if ( aTimeLeft < 0 ) {
                return "BROADCAST ENDED";
            }

            // Start the encoder if it is not already
            if ( itsEncoder == null ) {
                startEncoder(theChannelTime);
            }

            // If the timeleft is less than 10 seconds, change the repopulate value
            if ( aTimeLeft < 10*1000 ) {
                // This means that the repopulate will occur 5 seconds after the broadcast
                // concludes and the message changes to "BROADCAST ENDED"
                setRepopulateInMillis( aTimeLeft + 5000 );
            }

            rstr += getOnAirString(aTimeLeft);
            return rstr;
        }

        if ( aRealDelta > 0 ) {
            // If aRealDelta is within 20 seconds, disable repopulation
            // and get ready to start the encoder
            /* log("testing aRealDelta (" + aRealDelta
                + ") < BRM (" + BROADCAST_REQUEST_MILLIS + ") == " 
                + String.valueOf( aRealDelta < BROADCAST_REQUEST_MILLIS ) );
            */
            if ( aRealDelta < BROADCAST_REQUEST_MILLIS ) {
                // log("aRealDelta < BROADCAST_REQUEST_MILLIS, itsEncoder=" + itsEncoder);
                isRepopulateSuspended = true;
                if ( itsEncoder == null ) {
                    startEncoder(theChannelTime);
                }
            }

            rstr += Util.formatMillis(aRealDelta);
            return rstr;

        } else {
            // They are really on the air at this point
            theChannelTime.special = ChannelTime.ONAIR;
            return formatChannelTime(theChannelTime);
        }
    }

    private boolean startEncoder (ChannelTime theChannelTime) {

        log("startEncoder call starting");
        if ( itsEncoder == null ) {
            log("starting encoder...");
            try {
                itsEncoder = EncoderFactory.createEncoder(itsApp, theChannelTime);
                itsActiveChannel = theChannelTime;

            } catch ( Exception e ) {
                log("Could not create encoder: " + e.toString() 
                    + "\n" + Util.getStackTrace(e));
                return false;
            }
            Thread anEncoderThread = new Thread(itsEncoder);
            anEncoderThread.start();
            return true;

        } else {
            // huh?
            log("startEncoder called, but encoder was already broadcasting!");
            return false;
        }
    }

    public ChannelTime getActiveChannel () {
        return itsActiveChannel;
    }

    private void log ( String s ) {
        Main.log("[CHANNEL-UPDATE-THREAD] " + s);
    }

    private String getOnAirString( long aTimeLeft ) {
        if ( itsApp.getBroadcasting() ) {
            return "ON-AIR: " + Util.formatMillis(aTimeLeft) +  " remains.";
        } else {
            return "LATE! " + Util.formatMillis(aTimeLeft) +  " remains.";
        }
    }
}
