/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-client/src/org/opendj/client/Server.java,v 1.1.1.1 2001/09/29 20:22:00 jonathan Exp $
 * $Log: Server.java,v $
 * Revision 1.1.1.1  2001/09/29 20:22:00  jonathan
 * Imported opendj-client
 *
 */
package org.opendj.client;

import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.IOException;
import java.util.List;
import java.util.Vector;
import java.util.StringTokenizer;
import org.opendj.share.Util;
import org.opendj.share.ChannelTime;
import org.opendj.share.DjBean;
import org.opendj.share.GenreBean;
import org.opendj.share.MoodBean;

public class Server {

    public static final String CMD_AUTH               = "AUTH";
    public static final String CMD_LOGOUT             = "DISCONNECT";
    public static final String CMD_GETCHANNELTIMES    = "GETCHANNELTIMES";
    public static final String CMD_BROADCAST          = "BROADCAST";
    public static final String CMD_WAITFOREXIT        = "WAITFOREXIT";
    public static final String CMD_GETAUDIOSERVERINFO = "GETAUDIOSERVERINFO";
    public static final String CMD_GETSHOWSTATS       = "GETSHOWSTATS";
    public static final String CMD_GETGENRES          = "GETGENRES";
    public static final String CMD_GETMOODS           = "GETMOODS";
    public static final String CMD_UPDATESHOWMETADATA = "UPDATESHOWMETADATA";

    public static final String RES_AUTHOK              = "+AUTH";
    public static final String RES_NOSUCHUSER          = "+NOSUCHUSER";
    public static final String RES_BADPASSWORD         = "+BADPASSWORD";
    public static final String RES_GETCHANNELTIMESOK   = "+STARTCHANNELTIMES";
    public static final String RES_CHANNELTIME         = "+CHANNELTIME";
    public static final String RES_GETCHANNELTIMESDONE = "+ENDCHANNELTIMES";
    public static final String RES_LOGOUTOK            = "+BYE";
    public static final String RES_GO                  = "+GO";
    public static final String RES_DELAY               = "+DELAY";
    public static final String RES_CLOSENOW            = "+CLOSENOW";
    public static final String RES_EXPIRED             = "+EXPIRED";
    public static final String RES_AUDIOSERVER         = "+AUDIOSERVER";
    public static final String RES_SHOWSTATS           = "+SHOWSTATS";
    public static final String RES_GETGENRESOK         = "+STARTGENRES";
    public static final String RES_GENRE               = "+GENRE";
    public static final String RES_GETGENRESDONE       = "+ENDGENRES";
    public static final String RES_GETMOODSOK          = "+STARTMOODS";
    public static final String RES_MOOD                = "+MOOD";
    public static final String RES_GETMOODSDONE        = "+ENDMOODS";
    public static final String RES_METADATAUPDATED     = "+METADATAUPDATED";
    public static final String RES_FAIL                = "-";

    public static final int STATE_ERROR         = 0;
    public static final int STATE_DISCONNECTED  = 1;
    public static final int STATE_CONNECTED     = 2;
    public static final int STATE_AUTHENTICATED = 3;

    private int    itsState = STATE_DISCONNECTED;

    private String itsHost = null;
    private int    itsPort = -1;

    private Socket itsSocket = null;
    private BufferedReader itsInReader  = null;
    private PrintWriter    itsOutWriter = null;

    public Server (String theHost, int thePort) {
        itsHost = theHost;
        itsPort = thePort;
        itsState = STATE_DISCONNECTED;
    }

    private void connect () throws UnknownHostException, IOException {
        if ( itsState == STATE_DISCONNECTED ) {
            itsSocket = new Socket(itsHost, itsPort);
            itsInReader = new BufferedReader
                ( new InputStreamReader(itsSocket.getInputStream()));
            itsOutWriter = new PrintWriter
                ( new OutputStreamWriter(itsSocket.getOutputStream()), true);
            itsState = STATE_CONNECTED;
        } else {
            // Illegal state, just ignore it for now
        }
    }

    public DjBean authenticate ( String theEmail, 
                                 String thePassword ) 
        throws ServerException, UnknownHostException, IOException {

        String aServerResponse = null;
        DjBean aDjBean = null;

        if ( itsState == STATE_DISCONNECTED ) {
            connect();
        }

        if ( itsState != STATE_CONNECTED ) {
            return null;
        }

        String anAuthCommand = CMD_AUTH + " " + theEmail + " " + thePassword;
        synchronized (itsSocket) {
            log("sending AUTH command to server: '" + anAuthCommand + "'");
            sendData(anAuthCommand);
            log("command sent, getting read string");
            aServerResponse = readData();
        }
        log("got response: '" + aServerResponse + "'");
        
        if ( aServerResponse.startsWith(RES_AUTHOK) ) {
            // Authorized OK
            aDjBean = DjBean.parseServerLine(aServerResponse);
            itsState = STATE_AUTHENTICATED;
            return aDjBean;
        } else {
            if ( aServerResponse.startsWith(RES_NOSUCHUSER) ) {
                throw new ServerException(ServerException.E_NO_SUCH_USER);
            }
            if ( aServerResponse.startsWith(RES_BADPASSWORD) ) {
                throw new ServerException(ServerException.E_BADPASSWORD);
            }
            throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                      + ": " + aServerResponse);
        }
    }

    public void logout () throws ServerException, IOException {
        log("Server state: " + itsState);
        
        if ( itsState == STATE_AUTHENTICATED || itsState == STATE_CONNECTED ) {
            String aServerResponse = null;
            synchronized (itsSocket) {
                log("Sending DISCONNECT to server...");
                sendData(CMD_LOGOUT);
                log("Reading response from DISCONNECT...");
                aServerResponse = readData();
            }
                
            if ( aServerResponse.startsWith(RES_LOGOUTOK) ) {
                itsState = STATE_DISCONNECTED;
                try {itsSocket.close(); } catch ( Exception e2 ) {}
            } else {
                throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                          + ": " + aServerResponse);
            }
        }
    }

    public List getChannelTimes () throws ServerException, IOException {
        if ( itsState != STATE_AUTHENTICATED ) {
            return null;
        }

        List someChannelTimes = new Vector();
        String aServerResponse = null;
        ChannelTime aChannelTime = null;
        synchronized (itsSocket) {
            sendData(CMD_GETCHANNELTIMES);
            aServerResponse = readData();
            
            if ( aServerResponse.startsWith(RES_GETCHANNELTIMESOK) ) {
                long aRecvTime = System.currentTimeMillis();
                aServerResponse = readData();
                while ( aServerResponse.startsWith(RES_CHANNELTIME) ) {
                    aChannelTime = ChannelTime.parseServerLine(aServerResponse);
                    aChannelTime.recvTime = aRecvTime;
                    someChannelTimes.add( aChannelTime );
                    aServerResponse = readData();
                }
                if ( aServerResponse.startsWith(RES_GETCHANNELTIMESDONE) ) {
                    return someChannelTimes;
                } else {
                    throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                              + ": " + aServerResponse);
                }
            }
        }
        return null;
    }

    // This method blocks until permission to broadcast is granted.
    // Ideally, the server grants permission IMMEDIATELY after opening 
    // the hole in the firewall to let our encoder's packets through.
    // If the server denies access for any reason, this method will not
    // block but will return immediately.
    public boolean getPermissionToEncode (ChannelTime theChannelTime) 
        throws ServerException, IOException {

        // For testing
        // if ( true ) return true;

        String aServerResponse = null;
        String aCommand = null;
        synchronized (itsSocket) {
            sendData(CMD_BROADCAST + " " + theChannelTime.path);
            aServerResponse = readData();
        }
        
        if ( aServerResponse.startsWith(RES_GO)  ) {
            // Now let's get rolling!
            log("server says GO, returning true");
            return true;
            
        } else if ( aServerResponse.startsWith(RES_DELAY) ) {
            StringTokenizer st = new StringTokenizer(aServerResponse);
            st.nextToken(); // skip the command
            long aDelay = Long.parseLong(st.nextToken());
            
            // If the delay is less than 30 seconds, accept it and delay
            // but decrease the delay by 3 seconds so we don't miss our window
            // Otherwise return false
            if ( aDelay < (30*1000) ) {
                log("delay is less than 30 seconds, sleeping for (" 
                    + String.valueOf(aDelay - (3*1000)) + "), then asking again...");
                Util.sleep(Math.max(aDelay - (3*1000), 0));
                return getPermissionToEncode(theChannelTime);
            } else {
                // Delay is too long, give up
                log("delay is too long, returning false");
                return false;
            }
            
        } else {
            // This could be ALREADYBROADCASTING or NOTAUTHORIZED
            throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                      + ": " + aServerResponse);
        }
    }

    public long getWaitTime() throws ServerException, IOException {
        String aServerResponse = null;
        synchronized (itsSocket) {
            sendData(CMD_WAITFOREXIT);
            aServerResponse = readData();
        }
        
        if ( aServerResponse.startsWith(RES_CLOSENOW) ) {
            return 0;
        }
        if ( aServerResponse.startsWith(RES_DELAY) ) {
            StringTokenizer st = new StringTokenizer(aServerResponse);
            st.nextToken(); // skip the command
            long aDelay = Long.parseLong(st.nextToken());
            return aDelay;
        }
        if ( aServerResponse.startsWith(RES_EXPIRED) ) {
            // This means that the server has already disconnected us
            // No reason to continue broadcasting
            return 0;
        } else {
            throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                      + ": " + aServerResponse);
        }
    }

    public void loadAudioServerInfo(Main theApp) throws ServerException, IOException {
        String aServerResponse = null;
        synchronized (itsSocket) {
            sendData(CMD_GETAUDIOSERVERINFO);
            aServerResponse = readData();
        }
        
        if ( !aServerResponse.startsWith(RES_AUDIOSERVER) ) { 
            throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                      + ": " + aServerResponse);
        }
        
        StringTokenizer st = new StringTokenizer(aServerResponse);
        
        st.nextToken(); // Skip the response token
        String anIcecastHost     = st.nextToken();
        String anIcecastPort     = st.nextToken();
        String anIcecastPassword = st.nextToken();
        String aMaxBitrate       = st.nextToken();
        theApp.setProperty(Main.PROP_ICECAST_HOST,     anIcecastHost);
        theApp.setProperty(Main.PROP_ICECAST_PORT,     anIcecastPort);
        theApp.setProperty(Main.PROP_ICECAST_PASSWORD, anIcecastPassword);
        theApp.setMaxBitrate(Integer.parseInt(aMaxBitrate));
    }

    public OnAirStatus getOnAirStatus() throws ServerException, IOException {
        String aServerResponse = null;
        OnAirStatus rstat = null;
        synchronized (itsSocket) {
            sendData(CMD_GETSHOWSTATS);
            aServerResponse = readData();
        }
        if ( !aServerResponse.startsWith(RES_SHOWSTATS) ) {
            throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                      + ": " + aServerResponse);
        }
        
        StringTokenizer st = new StringTokenizer(aServerResponse);
        st.nextToken(); // Skip the response token
        rstat = new OnAirStatus();
        rstat.current = Integer.parseInt(st.nextToken());
        rstat.avg     = Integer.parseInt(st.nextToken());
        return rstat;
    }

    public List getGenreList () throws ServerException, IOException {
        if ( itsState != STATE_AUTHENTICATED ) {
            return null;
        }

        List aGenreList = new Vector();
        String aServerResponse = null;
        GenreBean aGenreBean = null;

        synchronized (itsSocket) {
            sendData(CMD_GETGENRES);
            aServerResponse = readData();
            
            if ( aServerResponse.startsWith(RES_GETGENRESOK) ) {
                aServerResponse = readData();
                while ( aServerResponse.startsWith(RES_GENRE) ) {
                    aGenreBean = GenreBean.parseServerLine(aServerResponse);
                    aGenreList.add( aGenreBean );
                    aServerResponse = readData();
                }
                if ( aServerResponse.startsWith(RES_GETGENRESDONE) ) {
                    return aGenreList;
                }
            } else {
                throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                          + ": " + aServerResponse);
            }
        }
        return null;
    }

    public List getMoodList () throws ServerException, IOException {
        if ( itsState != STATE_AUTHENTICATED ) {
            return null;
        }

        List aMoodList = new Vector();
        String aServerResponse = null;
        MoodBean aMoodBean = null;

        synchronized (itsSocket) {
            sendData(CMD_GETMOODS);
            aServerResponse = readData();
            
            if ( aServerResponse.startsWith(RES_GETMOODSOK) ) {
                aServerResponse = readData();
                while ( aServerResponse.startsWith(RES_MOOD) ) {
                    aMoodBean = MoodBean.parseServerLine(aServerResponse);
                    aMoodList.add( aMoodBean );
                    aServerResponse = readData();
                }
                if ( aServerResponse.startsWith(RES_GETMOODSDONE) ) {
                    return aMoodList;
                }
            } else {
                throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                          + ": " + aServerResponse);
            }
        }
        return null;
    }

    public void updateShowMetaData ( int theGenreID, 
                                     int theMoodID, 
                                     String theComments ) 
        throws ServerException, IOException {

        String aServerResponse = null;
        theComments = Util.stringReplace( theComments, "\n", "$__NL__$" );
        synchronized (itsSocket) {
            sendData(CMD_UPDATESHOWMETADATA + " |" + theGenreID + "|" + theMoodID + "|" + theComments);
            aServerResponse = readData();
        }
        if ( !aServerResponse.startsWith(RES_METADATAUPDATED) ) {
            throw new ServerException(ServerException.E_UNEXPECTED_RESPONSE 
                                      + ": " + aServerResponse);
        }
    }

    private void sendData ( String s ) {
        log(">" + s);
        itsOutWriter.println(s);
    }

    private String readData () throws IOException {
        String aReadData = itsInReader.readLine(); 
        log("<" + aReadData);
        return aReadData;
    }

    public InetAddress getRemoteIP() {
        return itsSocket.getInetAddress();
    }

    private void log ( String s ) {
        Main.log("[SERVER-SOCKET] " + s);
    }
}
