/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-client/src/org/opendj/client/java/EncoderImpl.java,v 1.1.1.1 2001/09/29 20:22:00 jonathan Exp $
 * $Log: EncoderImpl.java,v $
 * Revision 1.1.1.1  2001/09/29 20:22:00  jonathan
 * Imported opendj-client
 *
 */
package org.opendj.client.java;

import java.util.Properties;
import org.jrev.JRev;
import org.jrev.Logger;
import org.opendj.client.Main;
import org.opendj.client.Server;
import org.opendj.client.ServerException;
import org.opendj.client.Encoder;
import org.opendj.client.AlertDialog;
import org.opendj.share.Util;
import org.opendj.share.ChannelTime;
import org.opendj.share.DjBean;

public class EncoderImpl implements Encoder {

    public static final long SLEEP_INTERVAL = 10000;
    public static final long KILL_DELAY = 2000;
    public static final int DEFAULT_BITRATE = 24000;
    public static final boolean IS_PUBLIC      = true;
    public static final boolean NOT_PUBLIC     = false;
    public static final boolean ICY_HEADERS    = true;
    public static final boolean NO_ICY_HEADERS = false;

    public static final String PROP_JREV_LOGFILE = "jrev.logfile";

    private Main itsApp = null;
    private Server itsServer = null;
    private DjBean itsDjBean = null;
    private Properties itsProps = null;

    private ChannelTime itsChannelTime = null;
    private volatile boolean isDone = false;
    private JRev itsJRev = null;
    private Thread itsMainThread = null;
    private OpenDJPlaylist itsPlaylist = null;

    public EncoderImpl () {}

    public void init ( Main theApp, ChannelTime theChannelTime ) {
        log("EncoderImpl constructor called.");

        itsApp = theApp;
        itsServer = itsApp.getServer();
        itsProps = itsApp.getProps();
        itsDjBean = itsApp.getDjBean();
        itsChannelTime = theChannelTime;
        isDone = false;
    }

    public void run () {
        
        itsMainThread = Thread.currentThread();
        log("Encoder starting");

        try {
            int aBitRate = DEFAULT_BITRATE;
            try {
                aBitRate = Integer.parseInt(itsProps.getProperty(Main.PROP_USE_BITRATE));
            } catch ( Exception e ) {
                aBitRate = DEFAULT_BITRATE;
            }
            itsPlaylist = new OpenDJPlaylist(itsApp, itsChannelTime);

            log("Encoder is asking server for permission...");
            boolean hasPermission = false;
            try {
                hasPermission = itsServer.getPermissionToEncode(itsChannelTime);
            } catch ( ServerException se ) {
                AlertDialog a = new AlertDialog(itsApp, 
                                                AlertDialog.TITLE_ERROR,
                                                AlertDialog.MSG_ENCODE_PERMISSION_DENIED,
                                                se.getMessage());
                finish();
                return;
                
            } catch ( Exception e ) {
                AlertDialog a = new AlertDialog(itsApp, 
                                                AlertDialog.TITLE_ERROR,
                                                AlertDialog.MSG_ENCODE_PERMISSION_DENIED,
                                                e);
                finish();
                return;
            }

            if ( hasPermission ) {
                log("Server granted permission to encode");
            } else {
                log("Server did not grant permission to encode!");
                AlertDialog a = new AlertDialog(itsApp, 
                                                AlertDialog.TITLE_ERROR,
                                                AlertDialog.MSG_ENCODE_PERMISSION_DENIED);
                return;
            }

            itsJRev = new JRev( itsDjBean.getshowname(),
                                Util.EMPTY_STRING,
                                itsDjBean.getwebpage(),
                                itsDjBean.getgenreval(),
                                aBitRate,
                                itsProps.getProperty(Main.PROP_ICECAST_HOST),
                                Integer.parseInt(itsProps.getProperty(Main.PROP_ICECAST_PORT)),
                                itsProps.getProperty(Main.PROP_ICECAST_PASSWORD),
                                Util.stripUpTo(itsChannelTime.path, "/"),
                                NOT_PUBLIC,
                                ICY_HEADERS,
                                null,
                                null,
                                new Logger(false, itsProps.getProperty(PROP_JREV_LOGFILE), 999),
                                itsPlaylist );
            // itsJRev.start();
            while ( !isDone ) {
                Util.sleep(SLEEP_INTERVAL);
            }
        } finally {
            itsJRev.shutdown();
            itsPlaylist.finish();
        }
    }

    private void log ( String s ) {
        Main.log("[ENCODER-PROCESS-THREAD]: " + s);
    }

    public void finish() {
        isDone = true;
        itsMainThread.interrupt();
    }

    public boolean getIsDone() { return isDone; }
}
