/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-client/src/org/opendj/client/java/OpenDJStreamSource.java,v 1.1.1.1 2001/09/29 20:22:00 jonathan Exp $
 * $Log: OpenDJStreamSource.java,v $
 * Revision 1.1.1.1  2001/09/29 20:22:00  jonathan
 * Imported opendj-client
 *
 */
package org.opendj.client.java;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.ServerSocket;
import java.util.Properties;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.TargetDataLine;
import javax.sound.sampled.LineUnavailableException;
import org.jrev.StreamSource;
import org.opendj.client.Main;
import org.opendj.client.Server;
import org.opendj.client.AlertDialog;
import org.opendj.share.ChannelTime;
import org.opendj.share.Util;

public class OpenDJStreamSource implements StreamSource, Runnable {    

    public static final long SLEEP_INTERVAL = 50;
    public static final int MP3_SOCKET_PORT = 9026;
    
    public static final String[] ENCODING_NAMES 
        = { "ALAW", "PCM_SIGNED", "PCM_UNSIGNED", "ULAW" };
    public static final AudioFormat.Encoding[] ENCODING_VALUES
        = { AudioFormat.Encoding.ALAW,
            AudioFormat.Encoding.PCM_SIGNED,
            AudioFormat.Encoding.PCM_UNSIGNED,
            AudioFormat.Encoding.ULAW };

    public static final String PROP_ENCODING         = "jrev.encoding";
    public static final String PROP_SAMPLE_RATE      = "jrev.samplerate";
    public static final String PROP_SAMPLE_BITS      = "jrev.samplebits";
    public static final String PROP_NUM_CHANNELS     = "jrev.channels";
    public static final String PROP_FRAME_SIZE       = "jrev.framesize";
    public static final String PROP_FRAME_RATE       = "jrev.framerate";
    public static final String PROP_BIGENDIAN        = "jrev.bigendian";
    public static final String PROP_LINE_BUFFER_SIZE = "jrev.lineInBufferSize";
    public static final String PROP_LAME_BUFFER_SIZE = "jrev.lameBufferSize";
    public static final String PROP_LAME_COMMAND     = "jrev.lameCommand";

    private AudioFormat.Encoding itsEncoding       = AudioFormat.Encoding.PCM_UNSIGNED;
    private float                itsSampleRate     = 24000.0f;
    private int                  itsSampleBits     = 16;
    private int                  itsNumChannels    = 1;
    private int                  itsFrameSize      = 2;
    private float                itsFrameRate      = 22050.0f;
    private boolean              itsBigEndian      = true;
    private int                  itsLineBufferSize = 4096;
    private int                  itsLameBufferSize = 4096;
    private String               itsLameCommand    = null;
    private int                  itsBitrate        = -1;

    private volatile boolean isConnected         = false;
    private volatile boolean isServerSocketReady = false;
    private volatile boolean isDone              = false;

    private ServerSocket itsWriteSocket = null;
    private Socket       itsReadSocket  = null;

    private Main                  itsApp            = null;
    private Server                itsServer         = null;
    private ChannelTime           itsChannelTime    = null;
    private AudioFormat           itsAudioFormat    = null;
    private TargetDataLine        itsInputLine      = null;
    private Process               itsEncoderProcess = null;
    private OutputStream          itsOutStream      = null;
    private InputStream           itsInStream       = null;
    private InputStream           itsErrStream      = null;
    private InputStream           itsReadSocketInStream = null;
    private Thread                itsMainThread     = null;

    public OpenDJStreamSource (Main theApp, ChannelTime theChannelTime) {
        log("OpenDJStreamSource constructor called");
        itsApp = theApp;
        itsServer = theApp.getServer();
        itsChannelTime = theChannelTime;
        Properties aProps = theApp.getProps();

        String anEncodingProp = aProps.getProperty(PROP_ENCODING);
        int anIndex = Util.getStringIndex(anEncodingProp, ENCODING_NAMES);

        itsEncoding       = ENCODING_VALUES[anIndex];
        itsSampleRate     = Float.parseFloat(aProps.getProperty(PROP_SAMPLE_RATE));
        itsSampleBits     = Integer.parseInt(aProps.getProperty(PROP_SAMPLE_BITS));
        itsNumChannels    = Integer.parseInt(aProps.getProperty(PROP_NUM_CHANNELS));
        itsFrameSize      = Integer.parseInt(aProps.getProperty(PROP_FRAME_SIZE));
        itsFrameRate      = Float.parseFloat(aProps.getProperty(PROP_FRAME_RATE));
        itsBigEndian      = aProps.getProperty(PROP_BIGENDIAN).equals("true");
        itsLineBufferSize = Integer.parseInt(aProps.getProperty(PROP_LINE_BUFFER_SIZE));
        itsLameBufferSize = Integer.parseInt(aProps.getProperty(PROP_LAME_BUFFER_SIZE));
        itsLameCommand    = aProps.getProperty(PROP_LAME_COMMAND);
        itsBitrate        = Integer.parseInt(aProps.getProperty(Main.PROP_USE_BITRATE));
    }

    public int bitrate() { 
        log("OpenDJStreamSource.bitrate called");
        return itsBitrate;
    }

    public void close() {
        log("OpenDJStreamSource.close called");
        finish();
    }

    public void finish() { isDone = true; }

    public String name() { 
        log("OpenDJStreamSource.name called");
        return Util.EMPTY_STRING; 
    }

    public boolean open() { 

        // Build LAME command line 
        String aLameCommand
            = itsLameCommand + " -b " + String.valueOf(itsBitrate/1000) + " - - ";

        // Start the process
        try {
            itsEncoderProcess = Runtime.getRuntime().exec(aLameCommand);
        } catch ( Exception e ) {
            log("Error creating process: " + e.toString());
            log(Util.getStackTrace(e));
            AlertDialog a = new AlertDialog(itsApp, 
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ERROR_CREATING_ENCODER_PROCESS,
                                            e);
            return false;
        }

        // Get the streams for the process
        try {
            itsInStream  = itsEncoderProcess.getInputStream();
            itsErrStream = itsEncoderProcess.getErrorStream();
            itsOutStream = itsEncoderProcess.getOutputStream();
        } catch ( Exception e ) {
            log("Error creating/reading process streams: " + e.toString());
            log(Util.getStackTrace(e));
            AlertDialog a = new AlertDialog(itsApp, 
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ERROR_READING_ENCODER_STREAM, 
                                            e);
            return false;
        }

        log("OpenDJStreamSource.open called");
        itsAudioFormat = new AudioFormat( itsEncoding, 
                                          itsSampleRate,
                                          itsSampleBits,
                                          itsNumChannels,
                                          itsFrameSize,
                                          itsFrameRate,
                                          itsBigEndian );

        DataLine.Info info = new DataLine.Info(TargetDataLine.class, 
                                               itsAudioFormat);
        
        if (!AudioSystem.isLineSupported(info)) {
            log("OpenDJStreamSource.open: Line matching " + info + " not supported.");
            return false;
        }
        
        try {
            itsInputLine = (TargetDataLine) AudioSystem.getLine(info);
            itsInputLine.open(itsAudioFormat, itsInputLine.getBufferSize());

        } catch (LineUnavailableException ex) { 
            log("OpenDJStreamSource.open: Unable to open the line: " + ex);
            return false;

        } catch (SecurityException ex) { 
            log("OpenDJStreamSource.open: security violation: " + ex.toString());
            return false;

        } catch (Exception ex) { 
            log("OpenDJStreamSource.open: some other exception " + ex.toString());
            return false;
        }

        isDone = false;
        isConnected = false;

        // Setup the sockets
        try {
            itsWriteSocket = new ServerSocket(MP3_SOCKET_PORT);
        } catch ( Exception e ) {
            log("OpenDJStreamSource.open: could not setup server socket; " + e.toString());
            return false;
        }

        // Start pulling data from the line
        itsInputLine.start();
        log("OpenDJStreamSource.open: line started");

        // Start sending data to the encoder
        isServerSocketReady = false;
        Thread anEncoderThread = new Thread(this);
        anEncoderThread.start();

        // Wait for the server to be setup
        while ( !isServerSocketReady ) {
            Util.sleep(25);
        }

        itsApp.setBroadcasting(true);
        
        return true;
    }

    public float percentComplete() { 
        log("OpenDJStreamSource.percentComplete called");
        return 0.0f; 
    }

    public void run () {

        itsMainThread = Thread.currentThread();
        int aNumBytesRead = -1;

        byte[] aLameBuffer = new byte[itsLameBufferSize];

        // Get the socket connection
        Socket aReadSocket = null;
        OutputStream aReadSocketOutputStream = null;
        LineToEncoderWriter aLineToEncoderWriter = null;

        try {
            log("waiting for socket connection in accept()...");
            isServerSocketReady = true;
            aReadSocket = itsWriteSocket.accept();
            log("accept returned OK, socket connected");
            aReadSocketOutputStream = aReadSocket.getOutputStream();

            log("starting LineToEncoderWriter");
            aLineToEncoderWriter = new LineToEncoderWriter(itsInputLine, itsOutStream, itsLineBufferSize);
            Thread aLineToEncoderWriterThread = new Thread(aLineToEncoderWriter);
            aLineToEncoderWriterThread.start();

            while ( !isDone && aLineToEncoderWriter.isWriting() ) {
            
                // Grab the outputted data from the encoder
                aNumBytesRead = itsInStream.read(aLameBuffer, 0, aLameBuffer.length);

                if ( aNumBytesRead > 0 ) {
                    aReadSocketOutputStream.write(aLameBuffer, 0, aNumBytesRead);
                } else {
                    log("Nothing could be read from LAME!");
                }
            }

        } catch ( Exception e ) {
            log("OpenDJStreamSource.run: " + e.toString() + "\n" + Util.getStackTrace(e));
            return;

        } finally {
            // Stop the LineToEncoderWriter
            aLineToEncoderWriter.finish();

            // Stop the input line
            if ( itsInputLine != null ) {
                itsInputLine.stop();
                itsInputLine.close();
            }

            // Close the write socket
            try { itsWriteSocket.close(); } catch ( Exception e2 ) {}

            // Kill the encoder
            itsEncoderProcess.destroy();
        }
    }

    public int read ( byte[] buffer ) { 
        
        // log("OpenDJStreamSource.read called");
        if ( isDone ) return -1;

        try {
            if ( !isConnected ) connectSocket();

            return itsReadSocketInStream.read(buffer, 0, buffer.length);

        } catch ( Exception e ) {
            log("OpenDJStreamSource.read: " + e.toString());
            return -1;
        }
    }

    public String title() { 
        log("OpenDJStreamSource.title called");
        return Util.EMPTY_STRING; 
    }

    private void log ( String s ) {
        Main.log("[STREAM-SOURCE] " + s);
    }

    private void connectSocket() throws IOException {
        itsReadSocket = new Socket("127.0.0.1", MP3_SOCKET_PORT);
        itsReadSocketInStream = itsReadSocket.getInputStream();
        isConnected = true;
    }

    protected class LineToEncoderWriter implements Runnable {

        private TargetDataLine   itsInputLine        = null;
        private OutputStream     itsEncoderOutStream = null;
        private int              itsLineBufferSize   = -1;
        private volatile boolean isDone              = false;
        
        public LineToEncoderWriter ( TargetDataLine theInputLine, 
                                     OutputStream theEncoderOutStream,
                                     int theLineBufferSize ) {
            itsInputLine = theInputLine;
            itsEncoderOutStream = theEncoderOutStream;
            itsLineBufferSize = theLineBufferSize;
            isDone = false;
        }

        public void run () {

            int aNumBytesRead = -1;
            byte[] aLineBuffer = new byte[itsLineBufferSize];

            try {
                while ( !isDone ) {
                    // Read raw data from the line
                    aNumBytesRead = itsInputLine.read(aLineBuffer, 0, aLineBuffer.length);
                    if ( aNumBytesRead < 0 ) return;

                    // Write the raw data to the encoder
                    try {
                        // log("Writing " + aNumBytesRead + " to the encoder");
                        itsEncoderOutStream.write(aLineBuffer, 0, aNumBytesRead);
                    } catch ( Exception e ) {
                        log("LineToEncoderWriter.run: could not write to encoder: " + e.toString());
                        return;
                    }
                }
            } finally {
                finish();
            }
        }

        public boolean isWriting () { return !isDone; }

        public void finish() { isDone = true; }
    }
}
