/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-client/src/org/opendj/client/java/RawSoundTest.java,v 1.1.1.1 2001/09/29 20:22:00 jonathan Exp $
 * $Log: RawSoundTest.java,v $
 * Revision 1.1.1.1  2001/09/29 20:22:00  jonathan
 * Imported opendj-client
 *
 */
package org.opendj.client.java;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.TargetDataLine;
import javax.sound.sampled.LineUnavailableException;
import org.opendj.share.Util;

public class RawSoundTest {    

    public static final long SLEEP_INTERVAL = 50;
    
    public static final String[] ENCODING_NAMES 
        = { "ALAW", "PCM_SIGNED", "PCM_UNSIGNED", "ULAW" };
    public static final AudioFormat.Encoding[] ENCODING_VALUES
        = { AudioFormat.Encoding.ALAW,
            AudioFormat.Encoding.PCM_SIGNED,
            AudioFormat.Encoding.PCM_UNSIGNED,
            AudioFormat.Encoding.ULAW };

    public static final String PROP_ENCODING         = "jrev.encoding";
    public static final String PROP_SAMPLE_RATE      = "jrev.samplerate";
    public static final String PROP_SAMPLE_BITS      = "jrev.samplebits";
    public static final String PROP_NUM_CHANNELS     = "jrev.channels";
    public static final String PROP_FRAME_RATE       = "jrev.framerate";
    public static final String PROP_BIGENDIAN        = "jrev.bigendian";
    public static final String PROP_LINE_BUFFER_SIZE = "jrev.lineInBufferSize";
    public static final String PROP_TEST_SECONDS     = "test.seconds";

    private AudioFormat.Encoding itsEncoding       = AudioFormat.Encoding.PCM_UNSIGNED;
    private float                itsSampleRate     = 24000.0f;
    private int                  itsSampleBits     = 16;
    private int                  itsNumChannels    = 1;
    private float                itsFrameRate      = 22050.0f;
    private boolean              itsBigEndian      = true;
    private int                  itsLineBufferSize = 4096;
    private int                  itsTestSeconds    = 60;

    private volatile boolean isDone              = false;

    private AudioFormat           itsAudioFormat    = null;
    private TargetDataLine        itsInputLine      = null;

    public static void main ( String[] args ) {

        if ( args.length != 1 ) {
            printUsage();
            return;
        }

        Properties aProps = new Properties();
        FileInputStream aPropsInStream = null;

        try {
            aPropsInStream = new FileInputStream(args[0]);
            aProps.load(aPropsInStream);

        } catch ( Exception e ) {
            log("Could not load properties");

        } finally {
            try { aPropsInStream.close(); } catch ( Exception e2 ) {}
        }
        
        RawSoundTest t = new RawSoundTest(aProps);
        t.go();
    }

    public RawSoundTest (Properties theProps) {

        log("RawSoundTest constructor called");

        String anEncodingProp = theProps.getProperty(PROP_ENCODING);
        int anIndex = Util.getStringIndex(anEncodingProp, ENCODING_NAMES);

        itsEncoding       = ENCODING_VALUES[anIndex];
        itsSampleRate     = Float.parseFloat(theProps.getProperty(PROP_SAMPLE_RATE));
        itsSampleBits     = Integer.parseInt(theProps.getProperty(PROP_SAMPLE_BITS));
        itsNumChannels    = Integer.parseInt(theProps.getProperty(PROP_NUM_CHANNELS));
        itsFrameRate      = Float.parseFloat(theProps.getProperty(PROP_FRAME_RATE));
        itsBigEndian      = theProps.getProperty(PROP_BIGENDIAN).equals("true");
        itsLineBufferSize = Integer.parseInt(theProps.getProperty(PROP_LINE_BUFFER_SIZE));
        itsTestSeconds    = Integer.parseInt(theProps.getProperty(PROP_TEST_SECONDS));
    }

    public boolean go() { 

        log("RawSoundTest.open called");
        itsAudioFormat = new AudioFormat( itsEncoding, 
                                          itsSampleRate,
                                          itsSampleBits,
                                          itsNumChannels,
                                          (itsSampleBits/8)*itsNumChannels,
                                          itsFrameRate,
                                          itsBigEndian );

        DataLine.Info info = new DataLine.Info(TargetDataLine.class, 
                                               itsAudioFormat);
        
        if (!AudioSystem.isLineSupported(info)) {
            log("RawSoundTest.open: Line matching " + info + " not supported.");
            return false;
        }
        
        try {
            itsInputLine = (TargetDataLine) AudioSystem.getLine(info);
            itsInputLine.open(itsAudioFormat, itsInputLine.getBufferSize());

        } catch (LineUnavailableException ex) { 
            log("RawSoundTest.open: Unable to open the line: " + ex);
            return false;

        } catch (SecurityException ex) { 
            log("RawSoundTest.open: security violation: " + ex.toString());
            return false;

        } catch (Exception ex) { 
            log("RawSoundTest.open: some other exception " + ex.toString());
            return false;
        }

        // Start pulling data from the line
        itsInputLine.start();
        log("RawSoundTest.open: line started");

        isDone = false;
        int aNumBytesRead = -1;
        byte[] aLineBuffer = new byte[itsLineBufferSize];
        long aStartTime = System.currentTimeMillis();
        long anEndTime = aStartTime + (1000*itsTestSeconds);
        log("startTime=" + aStartTime + ",endTime=" + anEndTime);
        while ( !isDone ) {
            
            if ( System.currentTimeMillis() > anEndTime ) break;

            // Read raw data from the line
            aNumBytesRead = itsInputLine.read(aLineBuffer, 0, aLineBuffer.length);

            if ( aNumBytesRead < 0 ) break;

            // Write the raw data to stdout
            try {
                System.out.write(aLineBuffer, 0, aNumBytesRead);
            } catch ( Exception e ) {
                log("RawSoundTest.run: could not write data: " + e.toString());
                return false;
            }
        }

        log("RawSoundTest is done");

        finalize();
        
        return true;
    }

    public void finalize () {
        try { 
            itsInputLine.stop();
            itsInputLine.close();
        } catch ( Exception e ) {}
    }

    private static void log ( String s ) {
        System.err.println("[STREAM-SOURCE]: " + s);
    }

    private static void printUsage() {
        System.err.println("\n\nUsage: java " + RawSoundTest.class.getName() + " <prop-file> \n\n");
    }

}
