/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-client/src/org/opendj/client/unix/EncoderImpl.java,v 1.1.1.1 2001/09/29 20:22:00 jonathan Exp $
 * $Log: EncoderImpl.java,v $
 * Revision 1.1.1.1  2001/09/29 20:22:00  jonathan
 * Imported opendj-client
 *
 */
package org.opendj.client.unix;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Properties;
import org.opendj.client.Main;
import org.opendj.client.Server;
import org.opendj.client.ServerException;
import org.opendj.client.Encoder;
import org.opendj.client.AlertDialog;
import org.opendj.share.Util;
import org.opendj.share.ChannelTime;
import org.opendj.share.DjBean;

public class EncoderImpl implements Encoder {

    public static final long SLEEP_INTERVAL = 100;
    public static final long KILL_DELAY = 2000;
    public static final int READ_BUFFER_SIZE = 4096;

    private Main itsApp = null;
    private Server itsServer = null;
    private Properties itsProps = null;

    private String itsPathToLiveIce = null;
    private String itsPathToLiveIceKiller = null;
    private String itsPathToLiveIceConfigTemplate = null;
    private String itsPathToLiveIceConfig = null;

    private ChannelTime itsChannelTime = null;
    private volatile boolean isDone = false;
    private Process itsEncoderProcess = null;
    private Thread itsMainThread = null;

    public EncoderImpl () {}

    public void init ( Main theApp, ChannelTime theChannelTime ) {
        log("EncoderImpl constructor called.");
        itsApp = theApp;
        itsServer = itsApp.getServer();
        itsProps = itsApp.getProps();
        itsChannelTime = theChannelTime;
        isDone = false;
        itsPathToLiveIce = itsProps.getProperty("liveice.start.script");
        itsPathToLiveIceKiller = itsProps.getProperty("liveice.stop.script");
        itsPathToLiveIceConfigTemplate = itsProps.getProperty("liveice.config.template");
        itsPathToLiveIceConfig = itsProps.getProperty("liveice.config");
        itsEncoderProcess = null;
        itsMainThread = null;
    }

    public void run () {
        
        log("Encoder starting");
        itsMainThread = Thread.currentThread();

        // Generate the config file
        try {
            EncoderConfig.generateConfig(itsProps,
                                         itsApp.getDjBean(),
                                         itsChannelTime, 
                                         itsServer.getRemoteIP());
        } catch ( Exception e ) {
            log("Error creating encoder config");
            AlertDialog a = new AlertDialog(itsApp, 
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ERROR_CREATING_ENCODER_CONFIG,
                                            e);
            return;            
        }

        log("Encoder is asking server for permission...");
        boolean hasPermission = false;
        try {
            hasPermission = itsServer.getPermissionToEncode(itsChannelTime);
        } catch ( ServerException se ) {
            AlertDialog a = new AlertDialog(itsApp, 
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ENCODE_PERMISSION_DENIED,
                                            se.getMessage());
            finish();
            return;

        } catch ( Exception e ) {
            AlertDialog a = new AlertDialog(itsApp, 
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ENCODE_PERMISSION_DENIED,
                                            e);
            finish();
            return;
        }
        if ( hasPermission ) {
            log("Server granted permission to encode");
            itsApp.setBroadcasting(true);
        } else {
            log("Server did not grant permission to encode!");
            AlertDialog a = new AlertDialog(itsApp, 
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ENCODE_PERMISSION_DENIED);
            finish();
            return;
        }
        
        // Create an encoder Process
        try {
            itsEncoderProcess = Runtime.getRuntime().exec(itsPathToLiveIce);
        } catch ( Exception e ) {
            log("Error creating process: " + e.toString());
            log(Util.getStackTrace(e));
            AlertDialog a = new AlertDialog(itsApp,
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ERROR_CREATING_ENCODER_PROCESS,
                                            e);
            finish();
            return;
        }
        
        // Get the streams for the process
        try {
            while ( !isDone ) {
                
                try {
                    itsEncoderProcess.waitFor();
                } catch ( InterruptedException ie ) {}
                if ( !isDone ) {
                    log("Encoder Terminated Prematurely!  Check liveice.err for details.");
                    AlertDialog a = new AlertDialog(itsApp, 
                                                    AlertDialog.TITLE_ERROR,
                                                    AlertDialog.MSG_ERROR_ENCODER_TERMINATED_PREMATURELY);
                    isDone = true;
                    itsApp.setBroadcasting(false);
                }

            }
            killEncoder();
            
        } catch ( Exception e ) {
            log("Error creating/reading process streams: " + e.toString());
            log(Util.getStackTrace(e));
            AlertDialog a = new AlertDialog(itsApp, 
                                            AlertDialog.TITLE_ERROR,
                                            AlertDialog.MSG_ERROR_READING_ENCODER_STREAM,
                                            e);
            return;
        }
    }

    private void log ( String s ) {
        Main.log("[ENCODER-PROCESS-THREAD] " + s);
    }

    public void finish() {
        isDone = true;
        itsMainThread.interrupt();
        killEncoder();
    }

    private void killEncoder () {
        itsApp.setBroadcasting(false);
        try {
            log("killEncoder: killing encoder");
            if ( itsEncoderProcess != null ) {
                itsEncoderProcess.destroy();
            }
            Process p = Runtime.getRuntime().exec(itsPathToLiveIceKiller);
            p.waitFor();

        } catch ( Exception e ) {
            log("Error stopping/destryoing encoder process: " + e.toString());
            log(Util.getStackTrace(e));
        }
    }

    public boolean getIsDone() { return isDone; }
}
